// Steady state DCC accessory decoder for tortoise point motors.
// Mike Johnson 29 June 2023 MERG 2418

//#define TESTENV

#ifdef __16F628A
#pragma config FOSC = HS        // Oscillator Selection bits (HS oscillator: High-speed crystal/resonator on RA6/OSC2/CLKOUT and RA7/OSC1/CLKIN)
#pragma config WDTE = OFF       // Watchdog Timer Enable bit (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable bit (PWRT disabled)
#pragma config MCLRE = ON       // RA5/MCLR/VPP Pin Function Select bit (RA5/MCLR/VPP pin function is MCLR)
#pragma config BOREN = ON       // Brown-out Detect Enable bit (BOD enabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable bit (RB4/PGM pin has PGM function, low-voltage programming enabled)
#pragma config CPD = OFF        // Data EE Memory Code Protection bit (Data memory code protection off)
#pragma config CP = OFF         // Flash Program Memory Code Protection bit (Code protection off)
#endif

#include <xc.h>

#ifdef __16F628A
// Kit 53
// PIC16F628A with 8 Mhz Crystal or resonator

#define 	F2a  	PORTAbits.RA2		//  pin 1
#define 	F2b  	PORTAbits.RA3		//  pin 2
// spare            PORTAbits.RA4       //  pin 3 comparator outputs ? 
//			MCLR						//  pin 4 ICSP
//			Vss	0V						//  pin 5
#define 	F3a	    PORTBbits.RB0		//  pin 6
#define 	F3b  	PORTBbits.RB1		//  pin 7
#define 	F4a  	PORTBbits.RB2    	//  pin 8
#define 	F4b	    PORTBbits.RB3		//  pin 9
#ifdef TESTENV
#define 	datain  PORTBbits.RB4		//  pin 10
#define 	ackout  PORTBbits.RB5		//  pin 11
#else
#define 	ackout	PORTBbits.RB6		//  pin 12 ICSP PGC
#define 	datain  PORTBbits.RB7		//  pin 13 ICSP PGD
#endif
//			Vdd 5V						//  pin 14
//			Osc2						//  pin 15
//			Osc1						//  pin 16
#define		F1a		PORTAbits.RA0		//  pin 17
#define		F1b	    PORTAbits.RA1		//  pin 18
#endif

/*
  The following are NMRA (2012) defined CV's and ranges.
  They can all be accessed but most do nothing.

  NMRA
  CV0       not used 
  CV1	    Lower 6 address bits, 0 - 63, address = CV9 x 256 + CV1 * 4 + 1
  CV2	x	Auxiliary activation  -  not done
  CV3		F1 function 'on' time, 0 - 256, in 10ms steps. 0 is continuous.
  CV4		F2
  CV5		F3
  CV6		F4
  CV7		Version Number, 60
  CV8		Manufacturers ID, MERG 165
  CV9		Upper 3 bits of	address 0 - 7
  CV10-CV27 NMRA reserved
  CV28  x   Bidirectional comms - not done
  CV29	x	Configuration - not done
  CV33-CV39 For MERG use - not done

  CV7 and CV8 read only, write 8 to CV8 for factory reset 
*/

typedef unsigned char byte;

// interrupt data
volatile int bittime;
volatile byte prevhalfbit;
enum STATE {WAIT, PREAMBLE, BYTE };
volatile enum STATE state;
volatile byte bytes[6];
volatile byte bytecount, pktready;

// other data
byte cbytes[6], cbytecount; // buffer input packet
byte acktimer;
//const byte masks[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };
byte masks[8];
byte timeout[4];           // output timeouts
byte cvdata[40];           // cv data in ram

// CV data in EEPROM
__EEPROM_DATA(  0, 0, 0, 0, 0, 0, 0, 60);      // CV0 to 7
__EEPROM_DATA(165, 0, 0, 0, 0, 0, 0,  0);      // CV8 to 15
__EEPROM_DATA(  0, 0, 0, 0, 0, 0, 0,  0);      // CV16 to 23
__EEPROM_DATA(  0, 0, 0, 0, 0, 0, 0,  0);      // CV24 to 31
__EEPROM_DATA(  0, 0, 0, 0, 0, 0, 0,  0);      // CV32 to 39

/************************************************************/    
    
// portb change interrupt, use timer 1 to get interval
// DCC is 52us to 64us as half a 1 bit, 90us to 12ms for half a 0 bit.
// but the DCC signal may be either way up
// timer 1 is in 4uS steps

void __interrupt() __section("introutine") bitint(void)
{
   	static byte count, t;

    t = datain;              // read of PORTB clears interrupt flag.
	INTCONbits.RBIF = 0;     // clear interrupt flag
    bittime = (TMR1H << 8) + TMR1L; // 4 usec steps
	TMR1H = TMR1L = 0;
    
    if (bittime>=50/4 && bittime <=64/4) { // 1 bit
        if (prevhalfbit == 1)
            goto bit1in;
        prevhalfbit = 1;
    }
    else if (bittime>=90/4 && bittime<12000/4) { // 0 bit
        if (prevhalfbit == 0)
            goto bit0in;
        prevhalfbit = 0;
    }
    else { // error ?
        prevhalfbit = 99;
        state = WAIT;
    }
    return;

bit1in:
    prevhalfbit = 99;
    if (state==WAIT) {
        state = PREAMBLE;
        count = 0;
    }
    else if (state==PREAMBLE) {
		count++; // preamble count
    }
    else if (state==BYTE) {
		count++; // data bit count
		if (count<=8) {
			bytes[bytecount] <<= 1;
            bytes[bytecount] |= 1;
        }
		else { // 9th bit is 1 marks end of packet
            bytecount++;
            count = bytes[0] ^ bytes[1] ^ bytes[2];
            if (bytecount>3) {
                count ^= bytes[3];
                if (bytecount>4) {
                    count ^= bytes[4];
                    if (bytecount==6)
                        count ^= bytes[5];
                }
            }
            if (count==0) // valid error check
                pktready = 1;
            state = PREAMBLE;
		}
	}
    return;

bit0in:
    prevhalfbit = 99;
    if (state==PREAMBLE) {
        if (count<10)
            state = WAIT;
        else {
            bytecount = 0;
            state = BYTE;
            count = 0;
        }
    }
    else if (state==BYTE) {
		count++; // data bit count
		if (count<=8) {
			bytes[bytecount] <<= 1; // 0 shift in
        }
		else { // 9th bit is 0 between bytes in packet
			count = 0;
            bytecount++;
            if (bytecount>=6) // too many bytes
                state = WAIT;
		}
	}
} // end interrupt

/************************************************************/    

void startack(void) {
	ackout = 1;				// set ack line high
    acktimer = 4;           // in 2uS steps
}

void outputReversed(byte n) {
    timeout[n] = cvdata[n+3];   
    if (n==0) {
        F1a = 0;
        F1b = 1;
    }
    else if (n==1) {
        F2a = 0;
        F2b = 1;
    }
    else if (n==2) {
        F3a = 0;
        F3b = 1;
    }
    else {
        F4a = 0;
        F4b = 1;
    }
}

void outputNormal(byte n) {
    timeout[n] = cvdata[n+3];   
    if (n==0) {
        F1a = 1;
        F1b = 0;
    }
    else if (n==1) {
        F2a = 1;
        F2b = 0;
    }
    else if (n==2) {
        F3a = 1;
        F3b = 0;
    }
    else {
        F4a = 1;
        F4b = 0;
    }
}

void outputoff(byte n) {
    if (n==0) {
        F1a = 0;
        F1b = 0;
    }
    else if (n==1) {
        F2a = 0;
        F2b = 0;
    }
    else if (n==2) {
        F3a = 0;
        F3b = 0;
    }
    else {
        F4a = 0;
        F4b = 0;
    }
}

/*
void outputn(byte n) {
    F1a = n&0x01;
    F1b = (n>>1)&0x01;
    F2a = (n>>2)&0x01;
    F2b = (n>>3)&0x01;
    F3a = (n>>4)&0x01;
    F3b = (n>>5)&0x01;
    F4a = (n>>6)&0x01;
    F4b = (n>>7)&0x01;
}
*/

byte eepromread(byte a) {
    return EEPROM_READ(a);
}

void eepromwrite(byte a, byte v) {
    EEPROM_WRITE(a, v);
}

/************************************************************/    

void main(void) {
    static byte i, outputtimer;
    static int adr, cv;

#ifdef __16F628A
    PORTA = 0;              // datasheet 5.1 does this first
	CMCON = 0x07;   		// turn comparators OFF
	TRISA = 0x00;			// all outputs
#ifdef TESTENV
    TRISB = 0x10;
#else
	TRISB = 0x80;			// bit 7 input
#endif
	PORTA = 0x00; 			// initialise points off
	PORTB = 0x00;
	OPTION_REG = 0x80;      // no portb pullups, no TMR0 settings

	// timer 2, 2mS timer
	TMR2 = 0;
	T2CON = 0b00000111;     // 500nS * 16 = 8uS counts, 2mS for 256
    PR2 = 0xFF;             // TMR2IF every 2mS
    PIR1bits.TMR2IF = 0;

	// timer 1, dcc bit timer
	T1CON = 0b00110001;		// prescale 8, Fosc/4, enable timer
	TMR1L = TMR1H = 0;		// 16 bit time in 4 uS steps, TMR1L is 1mS max  
#endif
    
    // initialise interrupt variables    
	prevhalfbit = 99;
    state = WAIT;

    // initialise other data
    acktimer = 0;
    for (i=0; i<8; i++)
        masks[i] = (byte)(1 << i);
    
    outputtimer = 5;
    for (i=0; i<40; i++)
        cvdata[i] = eepromread(i);
    for (i=0; i<4; i++)
        outputNormal(i);

    INTCONbits.RBIF = 0;    // clear port b interrupt
	INTCONbits.RBIE = 1;	// enable b change interrupt
    INTCONbits.GIE = 1;     // enable interrupts
    
	// main loop
	while(1) {
        if (PIR1bits.TMR2IF==1) { // 2mS timer
            PIR1bits.TMR2IF = 0;
            
            if (acktimer!=0) { // 6mS ack pulse end
                acktimer--;
                if (acktimer==0) {
                    ackout = 0;
                }
            }
            
            outputtimer--;  // 10 msec steps
            if (outputtimer==0) {
                outputtimer = 5;
                for (i=0; i<4; i++) { // check output timer ending
                    if (timeout[i]!=0) {
                        timeout[i]--;
                        if (timeout[i]==0) {
                            outputoff(i);
                        }
                    }
                }
            }
        } // end 2mS timer
        
        if (pktready) { // DCC packet in
            cbytecount = bytecount;
            for (i=0; i<6; i++) // copy packet
                cbytes[i] = bytes[i];
            pktready = 0;
            
            // ops mode: Basic accessory decoder packet or broadcast
            // {preamble} 0 10AAAAAA 0 1AAACDDD 0 EEEEEEEE 1
            if ((cbytes[0] & 0xC0) == 0x80 && (cbytes[1]&0x80) == 0x80
              && (cbytecount == 3 || cbytecount == 6)) {
                adr = (cbytes[0]&0x3F) | (((~cbytes[1])&0x70)<<2); // device address
                if (adr == ((cvdata[1] & 0x3F) | (cvdata[9])<<6) | adr == 0x01ff) {
                    if ((cbytes[1]&0x01)==0)
                        outputReversed((cbytes[1]>>1)&0x03);
                    else
                        outputNormal((cbytes[1]>>1)&0x03);
                    startack();
                }
            }
           
            // service mode: configuration variable access 
            // {preamble} 0 0111CCVV 0 VVVVVVVV 0 DDDDDDDD 0 EEEEEEEE 1
            if ((cbytes[0]&0xF0)==0x70 && cbytecount == 4) {
                // CC = 01 Verify, = 11 Write, == 10 bit op
                cv = (((cbytes[0]&0x03)<<8) | cbytes[1]) + 1; // cv number
                if ((cbytes[0]&0xFC)==0x7C) { // write
                    if (cv>=1 && cv<40 && cv!=7 && cv!=8) {
                        cvdata[(byte)cv] = cbytes[2];
                        eepromwrite((byte)cv, cbytes[2]);
                        startack();
                    }
                    else if (cv==8 && cbytes[2]==8) { // factory reset
                        for (i=0; i<40; i++)
                            cvdata[i] = 0;
                        cvdata[7] = 60;
                        cvdata[8] = 165;
                        for (i=0; i<40; i++)
                            eepromwrite(i, cvdata[i]);
                        startack();
                    }
                } 
                else if ((cbytes[0]&0xFC)==0x74) { // Verify
                    if (cv>=1 && cv<40) {
                        if (cvdata[(byte)cv] == cbytes[2])
                            startack();
                    }
                }         
                else if ((cbytes[0]&0xFC)==0x78) { // Bit op
                    // cbytes[2] = 111CDBBB, C=1 write bit, c=0 verify bit
                    if ((cbytes[2]&0xF0) == 0xF0) { //write bit
                        if (cv>=1 && cv<=40 && cv!=7 && cv!=8) {
                            i = (byte)(((cbytes[2]&0x08)>>3)<<(cbytes[2]&0x07));
                            cvdata[cv] &= ~masks[cbytes[2]&0x07];
                            cvdata[cv] |= i;
                            eepromwrite((byte)cv, cvdata[cv]);
                            startack();
                        }
                    }
                    else if ((cbytes[2]&0xF0) == 0xE0){ // verify bit
                        if (cv>=1 && cv<=40) {
                            i = cvdata[(byte)cv];
                            i = (i >> (cbytes[2]&0x07)) & 0x01;
                            if (i == ((cbytes[2]>>3)&0x01))
                                startack();
                        }                    
                    }
                }         
            }
             
        } // end of DCC packet in
    } // end loop
} // end main